@@ -1,11 +1,11 @@
 /* This is a proposed C API for support of SASL
  *
- *********************************IMPORTANT******************************
- * send email to chris.newman@innosoft.com and sasl-bugs@andrew.cmu.edu *
- * if you need to add new error codes, callback types, property values, *
- * etc.   It is important to keep the multiple implementations of this  *
- * API from diverging.                                                  *
- *********************************IMPORTANT******************************
+ *********************************IMPORTANT*******************************
+ * send email to chris.newman@innosoft.com and cyrus-bugs@andrew.cmu.edu *
+ * if you need to add new error codes, callback types, property values,  *
+ * etc.   It is important to keep the multiple implementations of this   *
+ * API from diverging.                                                   *
+ *********************************IMPORTANT*******************************
  *
  * Basic Type Summary:
  *  sasl_conn_t       Context for a SASL connection negotiation
@@ -35,14 +35,13 @@
  *  sasl_getprop      Get property (e.g., user name, security layer info)
  *  sasl_setprop      Set property (e.g., external ssf)
  *  sasl_errdetail    Generate string from last error on connection
- *  sasl_usererr      Translate server error code to user error code
  *  sasl_errstring    Translate sasl error code to a string
  *  sasl_encode       Encode data to send using security layer
  *  sasl_decode       Decode data received using security layer
  *  
  * Utility functions:
- *  sasl_encode64     Encode data to send using security layer
- *  sasl_decode64     Decode data received using security layer
+ *  sasl_encode64     Encode data to send using MIME base64 encoding
+ *  sasl_decode64     Decode data received using MIME base64 encoding
  *  sasl_erasebuffer  Erase a buffer
  *
  * Client Function Summary:
@@ -64,7 +63,7 @@
  *  sasl_setpass      Change a password or add a user entry
  *  sasl_auxprop_request  Request auxiliary properties
  *  sasl_auxprop_getctx   Get auxiliary property context for connection
- *  sasl_auxprop_getnames Get auxiliary property values by name
+ *  sasl_auxprop_store    Store a set of auxiliary properties
  *
  * Basic client model:
  *  1. client calls sasl_client_init() at startup to load plug-ins
@@ -122,11 +121,10 @@
 #define SASL_H 1
 
 #define SASL_VERSION_MAJOR 2
-#define SASL_VERSION_MINOR 0
-#define SASL_VERSION_STEP 5
+#define SASL_VERSION_MINOR 1
+#define SASL_VERSION_STEP 16
 
 #include "prop.h"
-#define LIBSASL_API
 
 /*************
  * Basic API *
@@ -201,6 +199,9 @@ typedef struct sasl_secret {
  */
 typedef struct sasl_rand_s sasl_rand_t;
 
+#ifdef __cplusplus
+extern "C" {
+#endif
 
 /****************************
  * Configure Basic Services *
@@ -259,6 +260,7 @@ typedef unsigned sasl_ssf_t;
 /* usage flags provided to sasl_server_new and sasl_client_new:
  */
 #define SASL_SUCCESS_DATA    0x0004 /* server supports data on success */
+#define SASL_NEED_PROXY      0x0008 /* require a mech that allows proxying */
 
 /***************************
  * Security Property Types *
@@ -282,6 +284,8 @@ typedef unsigned sasl_ssf_t;
  * PASS_CREDENTIALS     -- require mechanisms which pass client
  *			   credentials, and allow mechanisms which can pass
  *			   credentials to do so
+ * MUTUAL_AUTH          -- require mechanisms which provide mutual
+ *			   authentication
  */
 #define SASL_SEC_NOPLAINTEXT      0x0001
 #define SASL_SEC_NOACTIVE         0x0002
@@ -289,7 +293,8 @@ typedef unsigned sasl_ssf_t;
 #define SASL_SEC_FORWARD_SECRECY  0x0008
 #define SASL_SEC_NOANONYMOUS      0x0010
 #define SASL_SEC_PASS_CREDENTIALS 0x0020
-#define SASL_SEC_MAXIUMUM         0x00FF
+#define SASL_SEC_MUTUAL_AUTH      0x0040
+#define SASL_SEC_MAXIMUM          0x00FF
 
 typedef struct sasl_security_properties 
 { 
@@ -445,7 +450,7 @@ typedef int sasl_verifyfile_t(void *context,
  * outputs:
  *  result        -- set to NUL terminated string
  *                   NULL = user cancel
- *  len           -- length of result, ignored with SASL_CB_SECRET
+ *  len           -- length of result
  * returns SASL_OK
  */
 typedef int sasl_getsimple_t(void *context, int id,
@@ -578,39 +583,46 @@ typedef int sasl_server_userdb_setpass_t(sasl_conn_t *conn,
  * User Canonicalization plugin is called.  It has the responsibility
  * of copying its output into the provided output buffers.
  * 
- *  user, ulen    -- user name (authorization name), may not be NUL terminated
- *                   may be same as out_user
- *  authid, alen  -- authentication name, may not be NUL terminated
- *                   may be same as out_authid
+ *  in, inlen     -- user name to canonicalize, may not be NUL terminated
+ *                   may be same buffer as out
  *  flags         -- not currently used, supplied by auth mechanism
  *  user_realm    -- the user realm (may be NULL in case of client)
- *  out_user      -- buffer to copy user name
- *  out_umax      -- max length of user name
- *  out_ulen      -- set to length of user name
- *  out_authid    -- buffer to copy authid to
- *  out_amax      -- max length of authid
- *  out_alen      -- set to length of authid (0 = same as user name)
+ *  out           -- buffer to copy user name
+ *  out_max       -- max length of user name
+ *  out_len       -- set to length of user name
  *
  * returns
  *  SASL_OK         on success
  *  SASL_BADPROT    username contains invalid character
  */
+
+/* User Canonicalization Function Flags */
+
+#define SASL_CU_NONE    0x00 /* Not a valid flag to pass */
+/* One of the following two is required */
+#define SASL_CU_AUTHID  0x01
+#define SASL_CU_AUTHZID 0x02
+
 typedef int sasl_canon_user_t(sasl_conn_t *conn,
 			      void *context,
-			      const char *user, unsigned ulen,
-			      const char *authid, unsigned alen,
+			      const char *in, unsigned inlen,
 			      unsigned flags,
 			      const char *user_realm,
-			      char *out_user,
-			      unsigned out_umax, unsigned *out_ulen,
-			      char *out_authid,
-			      unsigned out_amax, unsigned *out_alen);
+			      char *out,
+			      unsigned out_max, unsigned *out_len);
+
 #define SASL_CB_CANON_USER (0x8007)
 
 /**********************************
  * Common Client/server functions *
  **********************************/
 
+/* get sasl library version information
+ * implementation is a vendor-defined string
+ * version is a vender-defined representation of the version # */
+LIBSASL_API void sasl_version(const char **implementation,
+			      int *version);
+
 /* dispose of all SASL plugins.  Connection
  * states have to be disposed of before calling this.
  */
@@ -681,6 +693,7 @@ LIBSASL_API int sasl_getprop(sasl_conn_t *conn, int propnum,
                                  * unnecessary */
 #define SASL_MAXOUTBUF    2     /* security layer max output buf unsigned */  
 #define SASL_DEFUSERREALM 3	/* default realm passed to server_new */
+				/* or set with setprop */
 #define SASL_GETOPTCTX    4	/* context for getopt callback */
 #define SASL_CALLBACK     7	/* current callback function list */
 #define SASL_IPLOCALPORT  8	/* iplocalport string passed to server_new */
@@ -690,6 +703,7 @@ LIBSASL_API int sasl_getprop(sasl_conn_t *conn, int propnum,
 #define SASL_AUTHSOURCE   14	/* name of auth source last used, useful
 				 * for failed authentication tracking */
 #define SASL_MECHNAME     15    /* active mechanism name, if any */
+#define SASL_AUTHUSER     16    /* authentication/admin user */
 
 /* This returns a string which is either empty or has an error message
  * from sasl_seterror (e.g., from a plug-in or callback).  It differs
@@ -883,15 +897,23 @@ LIBSASL_API int sasl_server_new(const char *service,
 				unsigned flags,
 				sasl_conn_t **pconn);
 
+/* Return an array of NUL-terminated strings, terminated by a NULL pointer,
+ * which lists all possible mechanisms that the library can supply
+ *
+ * Returns NULL on failure. */
+LIBSASL_API const char ** sasl_global_listmech(void);
+
 /* This returns a list of mechanisms in a NUL-terminated string
+ *  conn          -- the connection to list mechanisms for (either client
+ *                   or server)
  *  user          -- restricts mechanisms to those available to that user
- *                   (may be NULL)
+ *                   (may be NULL, not used for client case)
  *  prefix        -- appended to beginning of result
  *  sep           -- appended between mechanisms
  *  suffix        -- appended to end of result
  * results:
  *  result        -- NUL terminated result which persists until next
- *                   call to sasl_listmech
+ *                   call to sasl_listmech for this sasl_conn_t
  *  plen          -- gets length of result (excluding NUL), may be NULL
  *  pcount        -- gets number of mechanisms, may be NULL
  *
@@ -1035,6 +1057,7 @@ LIBSASL_API int sasl_setpass(sasl_conn_t *conn,
 			     unsigned flags);
 #define SASL_SET_CREATE  0x01   /* create a new entry for user */
 #define SASL_SET_DISABLE 0x02	/* disable user account */
+#define SASL_SET_NOPLAIN 0x04	/* do not store secret in plain text */
 
 /*********************************************************
  * Auxiliary Property Support -- added by cjn 1999-09-29 *
@@ -1043,7 +1066,8 @@ LIBSASL_API int sasl_setpass(sasl_conn_t *conn,
 #define SASL_AUX_END      NULL	/* last auxiliary property */
 
 /* traditional Posix items (should be implemented on Posix systems) */
-#define SASL_AUX_PASSWORD "userPassword" /* User Password */
+#define SASL_AUX_PASSWORD_PROP "userPassword" /* User Password */
+#define SASL_AUX_PASSWORD "*" SASL_AUX_PASSWORD_PROP /* User Password (of authid) */
 #define SASL_AUX_UIDNUM   "uidNumber"	/* UID number for the user */
 #define SASL_AUX_GIDNUM   "gidNumber"	/* GID for the user */
 #define SASL_AUX_FULLNAME "gecos"	/* full name of the user, unix-style */
@@ -1086,6 +1110,25 @@ LIBSASL_API int sasl_auxprop_request(sasl_conn_t *conn,
  */
 LIBSASL_API struct propctx *sasl_auxprop_getctx(sasl_conn_t *conn);
 
+/* Store the set of auxiliary properties for the given user.
+ * Use functions in prop.h to set the content.
+ *
+ *  conn         connection context
+ *  ctx          property context from prop_new()/prop_request()/prop_set()
+ *  user         NUL terminated user
+ *
+ * Call with NULL 'ctx' to see if the backend allows storing properties.
+ *
+ * errors
+ *  SASL_OK       -- success
+ *  SASL_NOMECH   -- can not store some/all properties
+ *  SASL_BADPARAM -- bad conn/ctx/user parameter
+ *  SASL_NOMEM    -- out of memory
+ *  SASL_FAIL     -- failed to store
+ */
+LIBSASL_API int sasl_auxprop_store(sasl_conn_t *conn,
+				   struct propctx *ctx, const char *user);
+
 /**********************
  * security layer API *
  **********************/
@@ -1117,6 +1160,9 @@ LIBSASL_API int sasl_encodev(sasl_conn_t *conn,
 /* decode a block of data received using security layer
  *  returning the input buffer if there is no security layer.
  *  output is only valid until next call to sasl_decode
+ *
+ *  if outputlen is 0 on return, than the value of output is undefined.
+ *  
  * returns:
  *  SASL_OK      -- success (returns input if no layer negotiated)
  *  SASL_NOTDONE -- security layer negotiation not finished
@@ -1126,4 +1172,8 @@ LIBSASL_API int sasl_decode(sasl_conn_t *conn,
 			    const char *input, unsigned inputlen,
 			    const char **output, unsigned *outputlen);
 
+#ifdef __cplusplus
+}
+#endif
+
 #endif /* SASL_H */
