@@ -17,7 +17,11 @@
 #include "prop.h"
 #endif
 
-/* callback to lookup a sasl_callback_t for a connnection
+#ifdef __cplusplus
+extern "C" {
+#endif
+
+/* callback to lookup a sasl_callback_t for a connection
  * input:
  *  conn        -- the connection to lookup a callback for
  *  callbacknum -- the number of the callback
@@ -115,8 +119,9 @@ typedef struct sasl_utils {
 
     /* format a message and then pass it to the SASL_CB_LOG callback
      *
-     * use syslog()-style formatting (printf with %m as most recent errno
-     * error).  The implementation may use a fixed size buffer not smaller
+     * use syslog()-style formatting (printf with %m as a human readable text
+     * (strerror()) for the error specified as the parameter).
+     * The implementation may use a fixed size buffer not smaller
      * than 512 octets if it securely truncates the message.
      *
      * level is a SASL_LOG_* level (see sasl.h)
@@ -145,17 +150,18 @@ typedef struct sasl_utils {
     int (*prop_setvals)(struct propctx *ctx, const char *name,
 			const char **values);
     void (*prop_erase)(struct propctx *ctx, const char *name);
+    int (*auxprop_store)(sasl_conn_t *conn,
+			 struct propctx *ctx, const char *user);
 
     /* for additions which don't require a version upgrade; set to 0 */
     int (*spare_fptr1)();
     int (*spare_fptr2)();
-    int (*spare_fptr3)();
 } sasl_utils_t;
 
 /*
  * output parameters from SASL API
  *
- * created / destroyed by the glue code, though probabally filled in
+ * created / destroyed by the glue code, though probably filled in
  * by a combination of the plugin, the glue code, and the canon_user callback.
  *
  */
@@ -170,7 +176,9 @@ typedef struct sasl_out_params {
 
     /* security layer information */
     unsigned maxoutbuf;
-    sasl_ssf_t mech_ssf;
+    sasl_ssf_t mech_ssf;   /* Should be set non-zero if negotiation of a
+	 		    * security layer was *attempted*, even if
+			    * the negotiation failed */
     void *encode_context;
     int (*encode)(void *context, const struct iovec *invec, unsigned numiov,
 		  const char **output, unsigned *outputlen);
@@ -239,10 +247,8 @@ typedef struct sasl_client_params {
      *  added rjs3 2001-05-23
      *  Must be called once user name aquired if canon_user is non-NULL.
      *  conn        connection context
-     *  user        user name from wire protocol (need not be NUL terminated)
-     *  ulen        length of user name from wire protocol (0 = strlen(user))
-     *  authid      authid from wire protocol (need not be NUL terminated)
-     *  alen        length of authid from wire protocol (0 = strlen(authid))
+     *  in          user name from wire protocol (need not be NUL terminated)
+     *  len         length of user name from wire protocol (0 = strlen(user))
      *  flags       for SASL_CU_* flags
      *  oparams     the user, authid, ulen, alen, fields are
      *              set appropriately after canonicalization/copying and
@@ -261,8 +267,7 @@ typedef struct sasl_client_params {
      *  SASL_BADPROT  -- invalid user/authid
      */
     int (*canon_user)(sasl_conn_t *conn,
-                    const char *user, unsigned ulen,
-                    const char *authid, unsigned alen,
+                    const char *in, unsigned len,
                     unsigned flags,
                     sasl_out_params_t *oparams);
 
@@ -271,7 +276,9 @@ typedef struct sasl_client_params {
     int spare_int1;
     int spare_int2;
     int spare_int3;
-    int spare_int4;
+
+    /* flags field as passed to sasl_client_new */
+    unsigned flags;
 
     /* set to 0 initially, this allows a plugin with extended parameters
      * to work with an older framework by updating version as parameters
@@ -282,11 +289,29 @@ typedef struct sasl_client_params {
 
 /* features shared between client and server */
 /* These allow the glue code to handle client-first and server-last issues */
+
+/* This indicates that the mechanism prefers to do client-send-first
+ * if the protocol allows it. */
 #define SASL_FEAT_WANT_CLIENT_FIRST 0x0002
-#define SASL_FEAT_WANT_SERVER_LAST 0x0004
-/* This says that client-first is handled internally (e.g. DIGEST-MD5) */
-/* Note that this is not the preferred way of doing things */
-#define SASL_FEAT_INTERNAL_CLIENT_FIRST 0x0008
+
+/* This feature is deprecated, instead, plugins should set *serverout to
+ * non-NULL and return SASL_OK intelligently to allow flexible use of
+ * server-last semantics */
+/* #define SASL_FEAT_WANT_SERVER_LAST 0x0004 */
+
+/* This feature is deprecated, instead plugins should correctly set
+ * SASL_FEAT_SERVER_FIRST as needed */
+/* #define SASL_FEAT_INTERNAL_CLIENT_FIRST 0x0008 */
+
+/* This indicates that the plugin is server-first only. 
+ * Not defining either of SASL_FEAT_SERVER_FIRST or 
+ * SASL_FEAT_WANT_CLIENT_FIRST indicates that the mechanism will take care 
+ * of the client-first situation internally.
+ */
+#define SASL_FEAT_SERVER_FIRST 0x0010
+
+/* This plugin allows proxying */
+#define SASL_FEAT_ALLOWS_PROXY 0x0020
 
 /* client plug-in features */
 #define SASL_FEAT_NEEDSERVERFQDN 0x0001
@@ -307,7 +332,7 @@ typedef struct sasl_client_plug {
     unsigned features;
 
     /* required prompt ids, NULL = user/pass only */
-    const long *required_prompts;
+    const unsigned long *required_prompts;
     
     /* global state for mechanism */
     void *glob_context;
@@ -454,8 +479,6 @@ typedef struct sasl_server_params {
      *  conn        connection context
      *  user        user name from wire protocol (need not be NUL terminated)
      *  ulen        length of user name from wire protocol (0 = strlen(user))
-     *  authid      authid from wire protocol (need not be NUL terminated)
-     *  alen        length of authid from wire protocol (0 = strlen(authid))
      *  flags       for SASL_CU_* flags
      *  oparams     the user, authid, ulen, alen, fields are
      *              set appropriately after canonicalization/copying and
@@ -475,7 +498,6 @@ typedef struct sasl_server_params {
      */
     int (*canon_user)(sasl_conn_t *conn,
 		      const char *user, unsigned ulen,
-		      const char *authid, unsigned alen,
 		      unsigned flags,
 		      sasl_out_params_t *oparams);
     
@@ -501,8 +523,8 @@ typedef struct sasl_server_params {
     int spare_int2;
     int spare_int3;
 
-    /* secflags field as passed to sasl_server_new */
-    unsigned secflags;
+    /* flags field as passed to sasl_server_new */
+    unsigned flags;
 
     /* set to 0 initially, this allows a plugin with extended parameters
      * to work with an older framework by updating version as parameters
@@ -723,7 +745,7 @@ typedef struct sasl_canonuser {
     void *glob_context;
 
     /* name of plugin */
-    void *spare_ptr1;
+    char *name;
 
     /* free global state for plugin */
     void (*canon_user_free)(void *glob_context, const sasl_utils_t *utils);
@@ -732,16 +754,11 @@ typedef struct sasl_canonuser {
      *  glob_context     -- global context from this structure
      *  sparams          -- server params, note user_realm&propctx elements
      *  user             -- user to login as (may not be NUL terminated)
-     *  ulen             -- length of user name (0 = strlen(user))
-     *  authid           -- authentication name (may not be NUL terminated)
-     *  alen             -- length of authentication name (0 = strlen(authid))
+     *  len              -- length of user name (0 = strlen(user))
      *  flags            -- for SASL_CU_* flags
-     *  out_user         -- buffer to copy user name
-     *  out_umax         -- max length of user name
-     *  out_ulen         -- set to length of user name
-     *  out_authid       -- buffer to copy authid to
-     *  out_amax         -- max length of authid
-     *  out_alen         -- set to length of authid (0 = same as user name)
+     *  out              -- buffer to copy user name
+     *  out_max          -- max length of user name
+     *  out_len          -- set to length of user name
      *
      *  note that the output buffers MAY be the same as the input buffers.
      *
@@ -751,23 +768,17 @@ typedef struct sasl_canonuser {
      */
     int (*canon_user_server)(void *glob_context,
 			     sasl_server_params_t *sparams,
-			     const char *user, unsigned ulen,
-			     const char *authid, unsigned alen,
+			     const char *user, unsigned len,
 			     unsigned flags,
-			     char *out_user,
-			     unsigned out_umax, unsigned *out_ulen,
-			     char *out_authid,
-			     unsigned out_amax, unsigned *out_alen);
+			     char *out,
+			     unsigned out_umax, unsigned *out_ulen);
 
     int (*canon_user_client)(void *glob_context,
 			     sasl_client_params_t *cparams,
-			     const char *user, unsigned ulen,
-			     const char *authid, unsigned alen,
+			     const char *user, unsigned len,
 			     unsigned flags,
-			     char *out_user,
-			     unsigned out_umax, unsigned *out_ulen,
-			     char *out_authid,
-			     unsigned out_amax, unsigned *out_alen);
+			     char *out,
+			     unsigned out_max, unsigned *out_len);
 
     /* for additions which don't require a version upgrade; set to 0 */
     int (*spare_fptr1)();
@@ -775,7 +786,7 @@ typedef struct sasl_canonuser {
     int (*spare_fptr3)();
 } sasl_canonuser_plug_t;
 
-#define SASL_CANONUSER_PLUG_VERSION 4
+#define SASL_CANONUSER_PLUG_VERSION 5
 
 /* default name for canonuser plug-in entry point is "sasl_canonuser_init"
  *  similar to sasl_server_plug_init model, except only returns one
@@ -818,15 +829,31 @@ typedef struct sasl_auxprop_plug {
 			   unsigned flags,
 			   const char *user, unsigned ulen);
 
-    /* for additions which don't require a version upgrade; set to 0 */
-    void (*spare_fptr1)();
-    void (*spare_fptr2)();
+    /* name of the auxprop plugin */
+    char *name;
+
+    /* store the fields/values of an auxiliary property context (OPTIONAL)
+     *
+     * if ctx is NULL, just check if storing properties is enabled
+     *
+     * returns
+     *  SASL_OK         on success
+     *  SASL_FAIL       on failure
+     */
+    int (*auxprop_store)(void *glob_context,
+			 sasl_server_params_t *sparams,
+			 struct propctx *ctx,
+			 const char *user, unsigned ulen);
 } sasl_auxprop_plug_t;
 
 /* auxprop lookup flags */
 #define SASL_AUXPROP_OVERRIDE 0x01 /* if clear, ignore auxiliary properties
 				    * with non-zero len field.  If set,
 				    * override value of those properties */
+#define SASL_AUXPROP_AUTHZID  0x02 /* if clear, we are looking up the
+				    * authid flags (prefixed with *), otherwise
+				    * we are looking up the authzid flags
+				    * (no prefix) */
 
 #define SASL_AUXPROP_PLUG_VERSION 4
 
@@ -845,4 +872,8 @@ typedef int sasl_auxprop_init_t(const sasl_utils_t *utils,
 LIBSASL_API int sasl_auxprop_add_plugin(const char *plugname,
 					sasl_auxprop_init_t *auxpropfunc);
 
+#ifdef __cplusplus
+}
+#endif
+
 #endif /* SASLPLUG_H */
