@@ -1,9 +1,9 @@
 /* canonusr.c - user canonicalization support
  * Rob Siemborski
- * $Id: canonusr.c,v 1.8 2002/05/06 15:57:42 rjs3 Exp $
+ * $Id: canonusr.c,v 1.13 2003/09/30 18:38:34 rjs3 Exp $
  */
 /* 
- * Copyright (c) 2001 Carnegie Mellon University.  All rights reserved.
+ * Copyright (c) 1998-2003 Carnegie Mellon University.  All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
@@ -65,7 +65,7 @@ static canonuser_plug_list_t *canonuser_head = NULL;
  *                   null-terminate, and get into the outparams
  *
  *                   (handled by INTERNAL plugin) */
-/* Also does auxprop lookups once username is canonoicalized */
+/* Also does auxprop lookups once username is canonicalized */
 /* a zero ulen or alen indicates that it is strlen(value) */
 int _sasl_canon_user(sasl_conn_t *conn,
                      const char *user, unsigned ulen,
@@ -118,6 +118,7 @@ int _sasl_canon_user(sasl_conn_t *conn,
 
 	/* Point the input copy at the stored buffer */
 	user = user_buf;
+	ulen = *lenp;
     }
 
     /* which plugin are we supposed to use? */
@@ -287,7 +288,7 @@ static int _canonuser_internal(const sasl_utils_t *utils,
     begin_u = &(userin[i]);
     if(i>0) ulen -= i;
 
-    for(;isspace((int)begin_u[ulen-1]) && ulen > 0; ulen--);
+    for(;ulen > 0 && isspace((int)begin_u[ulen-1]); ulen--);
     if(begin_u == &(userin[ulen])) {
 	sasl_FREE(in_buf);
 	utils->seterror(utils->conn, 0, "All-whitespace username.");
@@ -302,15 +303,18 @@ static int _canonuser_internal(const sasl_utils_t *utils,
 	u_apprealm = strlen(sconn->user_realm) + 1;
     }
     
-    /* Now copy! (FIXME: check for SASL_BUFOVER?) */
+    /* Now Copy */
     memcpy(out_user, begin_u, MIN(ulen, out_umax));
     if(sconn && u_apprealm) {
+	if(ulen >= out_umax) return SASL_BUFOVER;
 	out_user[ulen] = '@';
 	memcpy(&(out_user[ulen+1]), sconn->user_realm,
 	       MIN(u_apprealm-1, out_umax-ulen-1));
     }
     out_user[MIN(ulen + u_apprealm,out_umax)] = '\0';
 
+    if(ulen + u_apprealm > out_umax) return SASL_BUFOVER;
+
     if(out_ulen) *out_ulen = MIN(ulen + u_apprealm,out_umax);
     
     sasl_FREE(in_buf);
