@@ -1,10 +1,10 @@
 /* SASL server API implementation
  * Rob Siemborski
  * Tim Martin
- * $Id: checkpw.c,v 1.53 2002/04/26 18:59:36 leg Exp $
+ * $Id: checkpw.c,v 1.66 2003/09/09 15:38:13 ken3 Exp $
  */
 /* 
- * Copyright (c) 2001 Carnegie Mellon University.  All rights reserved.
+ * Copyright (c) 1998-2003 Carnegie Mellon University.  All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
@@ -69,7 +69,6 @@
 #include <strings.h>
 #include <netdb.h>
 #include <netinet/in.h>
-#include <sys/param.h>
 #include <sys/un.h>
 #else
 #include <string.h>
@@ -93,8 +92,6 @@
 # ifdef HAVE_UNISTD_H
 #  include <unistd.h>
 # endif
-
-extern int errno;
 #endif
 
 
@@ -122,7 +119,7 @@ static int _sasl_make_plain_secret(const char *salt,
     _sasl_MD5Update(&ctx, "sasldb", 6);
     _sasl_MD5Update(&ctx, passwd, passlen);
     memcpy((*secret)->data, salt, 16);
-    memcpy((*secret)->data + 16, "\0", 1);
+    (*secret)->data[16] = '\0';
     _sasl_MD5Final((*secret)->data + 17, &ctx);
     (*secret)->len = sec_len;
     
@@ -143,7 +140,7 @@ static int auxprop_verify_password(sasl_conn_t *conn,
     int result = SASL_OK;
     sasl_server_conn_t *sconn = (sasl_server_conn_t *)conn;
     const char *password_request[] = { SASL_AUX_PASSWORD,
-				       "cmusaslsecretPLAIN",
+				       "*cmusaslsecretPLAIN",
 				       NULL };
     struct propval auxprop_values[3];
     
@@ -161,13 +158,10 @@ static int auxprop_verify_password(sasl_conn_t *conn,
     if(result != SASL_OK) return result;
 
     result = _sasl_canon_user(conn, userstr, 0,
-			      SASL_CU_AUTHID, &(conn->oparams));
+			      SASL_CU_AUTHID | SASL_CU_AUTHZID,
+			      &(conn->oparams));
     if(result != SASL_OK) return result;
     
-    result = _sasl_canon_user(conn, userstr, 0,
-			      SASL_CU_AUTHZID, &(conn->oparams));
-    if(result != SASL_OK) return result;
-
     result = prop_getnames(sconn->sparams->propctx, password_request,
 			   auxprop_values);
     if(result < 0)
@@ -241,7 +235,7 @@ int _sasl_auxprop_verify_apop(sasl_conn_t *conn,
     char *userid = NULL;
     char *realm = NULL;
     unsigned char digest[16];
-    char digeststr[32];
+    char digeststr[33];
     const char *password_request[] = { SASL_AUX_PASSWORD, NULL };
     struct propval auxprop_values[2];
     sasl_server_conn_t *sconn = (sasl_server_conn_t *)conn;
@@ -264,6 +258,7 @@ int _sasl_auxprop_verify_apop(sasl_conn_t *conn,
        !auxprop_values[0].values ||
        !auxprop_values[0].values[0]) {
 	sasl_seterror(conn, 0, "could not find password");
+	ret = SASL_NOUSER;
 	goto done;
     }
     
@@ -338,7 +333,7 @@ static int retry_writev(int fd, struct iovec *iov, int iovcnt)
 	written += n;
 
 	for (i = 0; i < iovcnt; i++) {
-	    if (iov[i].iov_len > (unsigned) n) {
+	    if ((int) iov[i].iov_len > n) {
 		iov[i].iov_base = (char *)iov[i].iov_base + n;
 		iov[i].iov_len -= n;
 		break;
@@ -367,10 +362,8 @@ static int pwcheck_verify_password(sasl_conn_t *conn,
     int r;
     struct iovec iov[10];
     static char response[1024];
-    int start, n;
+    unsigned start, n;
     char pwpath[1024];
-    sasl_getopt_t *getopt;
-    void *context;
 
     if (strlen(PWCHECKDIR)+8+1 > sizeof(pwpath)) return SASL_FAIL;
 
@@ -422,10 +415,11 @@ static int pwcheck_verify_password(sasl_conn_t *conn,
  * Keep calling the read() system call with 'fd', 'buf', and 'nbyte'
  * until all the data is read in or an error occurs.
  */
-static int retry_read(int fd, void *buf, unsigned nbyte)
+static int retry_read(int fd, void *buf0, unsigned nbyte)
 {
     int n;
     int nread = 0;
+    char *buf = buf0;
 
     if (nbyte == 0) return 0;
 
@@ -461,6 +455,7 @@ static int saslauthd_verify_password(sasl_conn_t *conn,
     void *context;
     char pwpath[sizeof(srvaddr.sun_path)];
     const char *p = NULL;
+    char *freeme = NULL;
 #ifdef USE_DOORS
     door_arg_t arg;
 #endif
@@ -479,6 +474,19 @@ static int saslauthd_verify_password(sasl_conn_t *conn,
 	strcat(pwpath, "/mux");
     }
 
+    /* Split out username/realm if necessary */
+    if(strrchr(userid,'@') != NULL) {
+	char *rtmp;
+	
+	if(_sasl_strdup(userid, &freeme, NULL) != SASL_OK)
+	    goto fail;
+
+	userid = freeme;
+	rtmp = strrchr(userid,'@');
+	*rtmp = '\0';
+	user_realm = rtmp + 1;
+    }
+
     /*
      * build request of the form:
      *
@@ -495,7 +503,7 @@ static int saslauthd_verify_password(sasl_conn_t *conn,
 	if (u_len + p_len + s_len + r_len + 30 > (unsigned short) sizeof(query)) {
 	    /* request just too damn big */
             sasl_seterror(conn, 0, "saslauthd request too large");
-	    return SASL_FAIL;
+	    goto fail;
 	}
 
 	u_len = htons(u_len);
@@ -523,8 +531,8 @@ static int saslauthd_verify_password(sasl_conn_t *conn,
 #ifdef USE_DOORS
     s = open(pwpath, O_RDONLY);
     if (s < 0) {
-	sasl_seterror(conn, 0, "cannot open door to saslauthd server: %m");
-	return SASL_FAIL;
+	sasl_seterror(conn, 0, "cannot open door to saslauthd server: %m", errno);
+	goto fail;
     }
 
     arg.data_ptr = query;
@@ -534,23 +542,28 @@ static int saslauthd_verify_password(sasl_conn_t *conn,
     arg.rbuf = response;
     arg.rsize = sizeof(response);
 
-    door_call(s, &arg);
+    if (door_call(s, &arg) < 0) {
+      /* Parameters are undefined */
+      sasl_seterror(conn, 0, "door call to saslauthd server failed: %m", errno);
+      goto fail;
+    }
 
     if (arg.data_ptr != response || arg.data_size >= sizeof(response)) {
 	/* oh damn, we got back a really long response */
 	munmap(arg.rbuf, arg.rsize);
 	sasl_seterror(conn, 0, "saslauthd sent an overly long response");
-	return SASL_FAIL;
+	goto fail;
     }
     response[arg.data_size] = '\0';
 
+    close(s);
 #else
     /* unix sockets */
 
     s = socket(AF_UNIX, SOCK_STREAM, 0);
     if (s == -1) {
-	sasl_seterror(conn, 0, "cannot create socket for saslauthd: %m");
-	return SASL_FAIL;
+	sasl_seterror(conn, 0, "cannot create socket for saslauthd: %m", errno);
+	goto fail;
     }
 
     memset((char *)&srvaddr, 0, sizeof(srvaddr));
@@ -560,8 +573,8 @@ static int saslauthd_verify_password(sasl_conn_t *conn,
     {
 	int r = connect(s, (struct sockaddr *) &srvaddr, sizeof(srvaddr));
 	if (r == -1) {
-	    sasl_seterror(conn, 0, "cannot connect to saslauthd server: %m");
-	    return SASL_FAIL;
+	    sasl_seterror(conn, 0, "cannot connect to saslauthd server: %m", errno);
+	    goto fail;
 	}
     }
 
@@ -573,7 +586,7 @@ static int saslauthd_verify_password(sasl_conn_t *conn,
 
 	if (retry_writev(s, iov, 1) == -1) {
             sasl_seterror(conn, 0, "write failed");
-  	    return SASL_FAIL;
+	    goto fail;
   	}
     }
 
@@ -587,21 +600,21 @@ static int saslauthd_verify_password(sasl_conn_t *conn,
 	 */
 	if (retry_read(s, &count, sizeof(count)) < (int) sizeof(count)) {
 	    sasl_seterror(conn, 0, "size read failed");
-	    return SASL_FAIL;
+	    goto fail;
 	}
 	
 	count = ntohs(count);
 	if (count < 2) { /* MUST have at least "OK" or "NO" */
 	    close(s);
 	    sasl_seterror(conn, 0, "bad response from saslauthd");
-	    return SASL_FAIL;
+	    goto fail;
 	}
 	
 	count = (int)sizeof(response) < count ? sizeof(response) : count;
 	if (retry_read(s, response, count) < count) {
 	    close(s);
 	    sasl_seterror(conn, 0, "read failed");
-	    return SASL_FAIL;
+	    goto fail;
 	}
 	response[count] = '\0';
     }
@@ -609,12 +622,18 @@ static int saslauthd_verify_password(sasl_conn_t *conn,
     close(s);
 #endif /* USE_DOORS */
   
+    if(freeme) free(freeme);
+
     if (!strncmp(response, "OK", 2)) {
 	return SASL_OK;
     }
   
     sasl_seterror(conn, SASL_NOLOG, "authentication failed");
     return SASL_BADAUTH;
+
+ fail:
+    if (freeme) free(freeme);
+    return SASL_FAIL;
 }
 
 #endif
