@@ -1,10 +1,10 @@
 /* common.c - Functions that are common to server and clinet
  * Rob Siemborski
  * Tim Martin
- * $Id: common.c,v 1.67 2001/12/04 02:05:25 rjs3 Exp $
+ * $Id: common.c,v 1.96 2003/09/02 15:34:02 rjs3 Exp $
  */
 /* 
- * Copyright (c) 2001 Carnegie Mellon University.  All rights reserved.
+ * Copyright (c) 1998-2003 Carnegie Mellon University.  All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
@@ -59,23 +59,24 @@
 #include <saslplug.h>
 #include "saslint.h"
 
-#ifdef WIN32
-/* need to handle the fact that errno has been defined as a function
-   in a dll, not an extern int */
-# ifdef errno
-#  undef errno
-# endif /* errno */
-#endif /* WIN32 */
 #ifdef HAVE_UNISTD_H
 #include <unistd.h>
 #endif
 
+static int _sasl_getpath(void *context __attribute__((unused)), const char **path);
+
 static const char build_ident[] = "$Build: libsasl " PACKAGE "-" VERSION " $";
 
+/* It turns out to be convenient to have a shared sasl_utils_t */
+LIBSASL_VAR const sasl_utils_t *sasl_global_utils = NULL;
+
+/* Should be a null-terminated array that lists the available mechanisms */
+static char **global_mech_list = NULL;
+
 void *free_mutex = NULL;
 
-void (*_sasl_client_cleanup_hook)(void) = NULL;
-void (*_sasl_server_cleanup_hook)(void) = NULL;
+int (*_sasl_client_cleanup_hook)(void) = NULL;
+int (*_sasl_server_cleanup_hook)(void) = NULL;
 int (*_sasl_client_idle_hook)(sasl_conn_t *conn) = NULL;
 int (*_sasl_server_idle_hook)(sasl_conn_t *conn) = NULL;
 
@@ -124,7 +125,7 @@ void sasl_set_mutex(sasl_mutex_alloc_t *n, sasl_mutex_lock_t *l,
 }
 
 /* copy a string to malloced memory */
-int _sasl_strdup(const char *in, char **out, int *outlen)
+int _sasl_strdup(const char *in, char **out, size_t *outlen)
 {
   size_t len = strlen(in);
   if (outlen) *outlen = len;
@@ -135,9 +136,10 @@ int _sasl_strdup(const char *in, char **out, int *outlen)
 }
 
 /* adds a string to the buffer; reallocing if need be */
-int _sasl_add_string(char **out, int *alloclen, int *outlen, const char *add)
+int _sasl_add_string(char **out, size_t *alloclen,
+		     size_t *outlen, const char *add)
 {
-  int addlen;
+  size_t addlen;
 
   if (add==NULL) add = "(null)";
 
@@ -151,6 +153,18 @@ int _sasl_add_string(char **out, int *alloclen, int *outlen, const char *add)
   return SASL_OK;
 }
 
+/* return the version of the cyrus sasl library as compiled,
+ * using 32 bits: high byte is major version, second byte is minor version,
+ * low 16 bits are step # */
+void sasl_version(const char **implementation, int *version) 
+{
+    const char *implementation_string = "Cyrus SASL";
+    if(implementation) *implementation = implementation_string;
+    if(version) *version = (SASL_VERSION_MAJOR << 24) | 
+		           (SASL_VERSION_MINOR << 16) |
+		           (SASL_VERSION_STEP);
+}
+
 /* security-encode a regular string.  Mostly a wrapper for sasl_encodev */
 /* output is only valid until next call to sasl_encode or sasl_encodev */
 int sasl_encode(sasl_conn_t *conn, const char *input,
@@ -236,21 +250,27 @@ int sasl_decode(sasl_conn_t *conn,
     if(!conn->props.maxbufsize) {
 	sasl_seterror(conn, 0,
 		      "called sasl_decode with application that does not support security layers");
-	return SASL_TOOWEAK;
+	RETURN(conn, SASL_TOOWEAK);
     }
 
-    if(inputlen > conn->oparams.maxoutbuf)
-	PARAMERROR(conn);
-    
     if(conn->oparams.decode == NULL)
     {
 	/* Since we know how long the output is maximally, we can
 	 * just allocate it to begin with, and never need another
          * allocation! */
+
+	/* However, if they pass us more than they actually can take,
+	 * we cannot help them... */
+	if(inputlen > conn->props.maxbufsize) {
+	    sasl_seterror(conn, 0,
+			  "input too large for default sasl_decode");
+	    RETURN(conn,SASL_BUFOVER);
+	}
+
 	if(!conn->decode_buf)
-	    conn->decode_buf = sasl_ALLOC(conn->oparams.maxoutbuf + 1);
-	if(!conn->decode_buf)
-	    MEMERROR( conn );
+	    conn->decode_buf = sasl_ALLOC(conn->props.maxbufsize + 1);
+	if(!conn->decode_buf)	
+	    MEMERROR(conn);
 	
 	memcpy(conn->decode_buf, input, inputlen);
 	conn->decode_buf[inputlen] = '\0';
@@ -261,6 +281,10 @@ int sasl_decode(sasl_conn_t *conn,
     } else {
         result = conn->oparams.decode(conn->context, input, inputlen,
                                       output, outputlen);
+
+	/* NULL an empty buffer (for misbehaved applications) */
+	if (*outputlen == 0) *output = NULL;
+
         RETURN(conn, result);
     }
 
@@ -282,32 +306,35 @@ sasl_set_alloc(sasl_malloc_t *m,
 
 void sasl_done(void)
 {
-  if (_sasl_server_cleanup_hook)
-    _sasl_server_cleanup_hook();
-
-  if (_sasl_client_cleanup_hook)
-    _sasl_client_cleanup_hook();
-  
-  _sasl_canonuser_free();
-  _sasl_done_with_plugins();
-  
-  sasl_MUTEX_FREE(free_mutex);
-  free_mutex = NULL;
-
-  _sasl_free_utils(&sasl_global_utils);
-
-  /* in case of another init/done */
-  _sasl_server_cleanup_hook = NULL;
-  _sasl_client_cleanup_hook = NULL;
-
-  _sasl_client_idle_hook = NULL;
-  _sasl_server_idle_hook = NULL;
+    if (_sasl_server_cleanup_hook && _sasl_server_cleanup_hook() == SASL_OK) {
+	_sasl_server_idle_hook = NULL;
+	_sasl_server_cleanup_hook = NULL;
+    }
+    
+    if (_sasl_client_cleanup_hook && _sasl_client_cleanup_hook() == SASL_OK) {
+	_sasl_client_idle_hook = NULL;	
+	_sasl_client_cleanup_hook = NULL;
+    }
+    
+    if(_sasl_server_cleanup_hook || _sasl_client_cleanup_hook)
+	return;
+    
+    _sasl_canonuser_free();
+    _sasl_done_with_plugins();
+    
+    sasl_MUTEX_FREE(free_mutex);
+    free_mutex = NULL;
+    
+    _sasl_free_utils(&sasl_global_utils);
+    
+    if(global_mech_list) sasl_FREE(global_mech_list);
+    global_mech_list = NULL;
 }
 
 /* fills in the base sasl_conn_t info */
 int _sasl_conn_init(sasl_conn_t *conn,
 		    const char *service,
-		    int flags,
+		    unsigned int flags,
 		    enum Sasl_conn_type type,
 		    int (*idle_hook)(sasl_conn_t *conn),
 		    const char *serverFQDN,
@@ -359,7 +386,6 @@ int _sasl_conn_init(sasl_conn_t *conn,
   conn->errdetail_buf[0] = '\0';
   
   conn->decode_buf = NULL;
-  conn->decode_buf_len = 0;
 
   if(serverFQDN) {
       result = _sasl_strdup(serverFQDN, &conn->serverFQDN, NULL);
@@ -380,16 +406,13 @@ int _sasl_conn_init(sasl_conn_t *conn,
   RETURN(conn, SASL_OK);
 }
 
-/* It turns out to be conveinent to have a shared sasl_utils_t */
-const sasl_utils_t *sasl_global_utils = NULL;
-
-int _sasl_common_init(void)
+int _sasl_common_init(sasl_global_callbacks_t *global_callbacks)
 {
     int result;
     
     /* Setup the global utilities */
     if(!sasl_global_utils) {
-	sasl_global_utils = _sasl_alloc_utils(NULL, NULL);
+	sasl_global_utils = _sasl_alloc_utils(NULL, global_callbacks);
 	if(sasl_global_utils == NULL) return SASL_NOMEM;
     }
 
@@ -450,6 +473,9 @@ void _sasl_conn_dispose(sasl_conn_t *conn) {
   if(conn->decode_buf)
       sasl_FREE(conn->decode_buf);
 
+  if(conn->mechlist_buf)
+      sasl_FREE(conn->mechlist_buf);
+
   if(conn->service)
       sasl_FREE(conn->service);
 
@@ -514,6 +540,12 @@ int sasl_getprop(sasl_conn_t *conn, int propnum, const void **pvalue)
       else
 	  *((const char **)pvalue) = conn->oparams.user;
       break;
+  case SASL_AUTHUSER:
+      if(! conn->oparams.authid)
+	  result = SASL_NOTDONE;
+      else
+	  *((const char **)pvalue) = conn->oparams.authid;
+      break;
   case SASL_SERVERFQDN:
       *((const char **)pvalue) = conn->serverFQDN;
       break;
@@ -527,9 +559,17 @@ int sasl_getprop(sasl_conn_t *conn, int propnum, const void **pvalue)
       break;
   case SASL_AUTHSOURCE: /* name of plugin (not name of mech) */
       if(conn->type == SASL_CONN_CLIENT) {
+	  if(!((sasl_client_conn_t *)conn)->mech) {
+	      result = SASL_NOTDONE;
+	      break;
+	  }
 	  *((const char **)pvalue) =
 	      ((sasl_client_conn_t *)conn)->mech->plugname;
       } else if (conn->type == SASL_CONN_SERVER) {
+	  if(!((sasl_server_conn_t *)conn)->mech) {
+	      result = SASL_NOTDONE;
+	      break;
+	  }
 	  *((const char **)pvalue) =
 	      ((sasl_server_conn_t *)conn)->mech->plugname;
       } else {
@@ -538,9 +578,17 @@ int sasl_getprop(sasl_conn_t *conn, int propnum, const void **pvalue)
       break;
   case SASL_MECHNAME: /* name of mech */
       if(conn->type == SASL_CONN_CLIENT) {
+	  if(!((sasl_client_conn_t *)conn)->mech) {
+	      result = SASL_NOTDONE;
+	      break;
+	  }
 	  *((const char **)pvalue) =
 	      ((sasl_client_conn_t *)conn)->mech->plug->mech_name;
       } else if (conn->type == SASL_CONN_SERVER) {
+	  if(!((sasl_server_conn_t *)conn)->mech) {
+	      result = SASL_NOTDONE;
+	      break;
+	  }
 	  *((const char **)pvalue) =
 	      ((sasl_server_conn_t *)conn)->mech->plug->mech_name;
       } else {
@@ -567,6 +615,10 @@ int sasl_getprop(sasl_conn_t *conn, int propnum, const void **pvalue)
 
   if(result == SASL_BADPARAM) {
       PARAMERROR(conn);
+  } else if(result == SASL_NOTDONE) {
+      sasl_seterror(conn, SASL_NOLOG,
+		    "Information that was requested is not yet available.");
+      RETURN(conn, result);
   } else if(result != SASL_OK) {
       INTERROR(conn, result);
   } else
@@ -591,6 +643,13 @@ int sasl_setprop(sasl_conn_t *conn, int propnum, const void *value)
   {
   case SASL_SSF_EXTERNAL:
       conn->external.ssf = *((sasl_ssf_t *)value);
+      if(conn->type == SASL_CONN_SERVER) {
+	((sasl_server_conn_t*)conn)->sparams->external_ssf =
+	  conn->external.ssf;
+      } else {
+	((sasl_client_conn_t*)conn)->cparams->external_ssf =
+	  conn->external.ssf;
+      }
       break;
 
   case SASL_AUTH_EXTERNAL:
@@ -608,11 +667,49 @@ int sasl_setprop(sasl_conn_t *conn, int propnum, const void *value)
 
       break;
 
-  case SASL_SEC_PROPS:
-      memcpy(&(conn->props),(sasl_security_properties_t *)value,
-	     sizeof(sasl_security_properties_t));
+  case SASL_DEFUSERREALM:
+      if(conn->type != SASL_CONN_SERVER) {
+	sasl_seterror(conn, 0, "Tried to set realm on non-server connection");
+	result = SASL_BADPROT;
+	break;
+      }
+
+      if(value && strlen(value)) {
+	  result = _sasl_strdup(value, &str, NULL);
+	  if(result != SASL_OK) MEMERROR(conn);
+      } else {
+	  PARAMERROR(conn);
+      }
+
+      if(((sasl_server_conn_t *)conn)->user_realm)
+      	  sasl_FREE(((sasl_server_conn_t *)conn)->user_realm);
+
+      ((sasl_server_conn_t *)conn)->user_realm = str;
+      ((sasl_server_conn_t *)conn)->sparams->user_realm = str;
+
       break;
 
+  case SASL_SEC_PROPS:
+  {
+      sasl_security_properties_t *props = (sasl_security_properties_t *)value;
+
+      if(props->maxbufsize == 0 && props->min_ssf != 0) {
+	  sasl_seterror(conn, 0,
+			"Attempt to disable security layers (maxoutbuf == 0) with min_ssf > 0");
+	  RETURN(conn, SASL_TOOWEAK);
+      }
+
+      conn->props = *props;
+
+      if(conn->type == SASL_CONN_SERVER) {
+	((sasl_server_conn_t*)conn)->sparams->props = *props;
+      } else {
+	((sasl_client_conn_t*)conn)->cparams->props = *props;
+      }
+
+      break;
+  }
+      
   case SASL_IPREMOTEPORT:
   {
       const char *ipremoteport = (const char *)value;
@@ -631,17 +728,23 @@ int sasl_setprop(sasl_conn_t *conn, int propnum, const void *value)
 	  if(conn->type == SASL_CONN_CLIENT) {
 	      ((sasl_client_conn_t *)conn)->cparams->ipremoteport
 		  = conn->ipremoteport;
+	      ((sasl_client_conn_t *)conn)->cparams->ipremlen =
+		  strlen(conn->ipremoteport);
 	  } else if (conn->type == SASL_CONN_SERVER) {
 	      ((sasl_server_conn_t *)conn)->sparams->ipremoteport
 		  = conn->ipremoteport;
+	      ((sasl_server_conn_t *)conn)->sparams->ipremlen =
+		  strlen(conn->ipremoteport);
 	  }
       } else {
 	  if(conn->type == SASL_CONN_CLIENT) {
 	      ((sasl_client_conn_t *)conn)->cparams->ipremoteport
 		  = NULL;
+	      ((sasl_client_conn_t *)conn)->cparams->ipremlen = 0;
 	  } else if (conn->type == SASL_CONN_SERVER) {
 	      ((sasl_server_conn_t *)conn)->sparams->ipremoteport
-		  = NULL;
+		  = NULL;	      
+	      ((sasl_server_conn_t *)conn)->sparams->ipremlen = 0;
 	  }
       }
 
@@ -666,17 +769,23 @@ int sasl_setprop(sasl_conn_t *conn, int propnum, const void *value)
 	  if(conn->type == SASL_CONN_CLIENT) {
 	      ((sasl_client_conn_t *)conn)->cparams->iplocalport
 		  = conn->iplocalport;
+	      ((sasl_client_conn_t *)conn)->cparams->iploclen
+		  = strlen(conn->iplocalport);
 	  } else if (conn->type == SASL_CONN_SERVER) {
 	      ((sasl_server_conn_t *)conn)->sparams->iplocalport
 		  = conn->iplocalport;
+	      ((sasl_server_conn_t *)conn)->sparams->iploclen
+		  = strlen(conn->iplocalport);
 	  }
       } else {
 	  if(conn->type == SASL_CONN_CLIENT) {
 	      ((sasl_client_conn_t *)conn)->cparams->iplocalport
 		  = NULL;
+	      ((sasl_client_conn_t *)conn)->cparams->iploclen = 0;
 	  } else if (conn->type == SASL_CONN_SERVER) {
 	      ((sasl_server_conn_t *)conn)->sparams->iplocalport
 		  = NULL;
+	      ((sasl_server_conn_t *)conn)->sparams->iploclen = 0;
 	  }
       }
       break;
@@ -754,6 +863,8 @@ const char *sasl_errdetail(sasl_conn_t *conn)
     unsigned need_len;
     const char *errstr;
     char leader[128];
+
+    if(!conn) return NULL;
     
     errstr = sasl_errstring(conn->error_code, NULL, NULL);
     snprintf(leader,128,"SASL(%d): %s: ",
@@ -768,12 +879,14 @@ const char *sasl_errdetail(sasl_conn_t *conn)
 }
 
 
-static int
-_sasl_global_getopt(void *context,
-		    const char *plugin_name,
-		    const char *option,
-		    const char ** result,
-		    unsigned *len)
+/* Note that this needs the global callbacks, so if you don't give getcallbacks
+ * a sasl_conn_t, you're going to need to pass it yourself (or else we couldn't
+ * have client and server at the same time */
+static int _sasl_global_getopt(void *context,
+			       const char *plugin_name,
+			       const char *option,
+			       const char ** result,
+			       unsigned *len)
 {
   const sasl_global_callbacks_t * global_callbacks;
   const sasl_callback_t *callback;
@@ -784,17 +897,19 @@ _sasl_global_getopt(void *context,
       for (callback = global_callbacks->callbacks;
 	   callback->id != SASL_CB_LIST_END;
 	   callback++) {
-	  if (callback->id == SASL_CB_GETOPT
-	      && (((sasl_getopt_t *)(callback->proc))(callback->context,
-						      plugin_name,
-						      option,
-						      result,
-						      len)
-		  == SASL_OK))
-	      return SASL_OK;
+	if (callback->id == SASL_CB_GETOPT) {
+	  if (!callback->proc) return SASL_FAIL;
+	  if (((sasl_getopt_t *)(callback->proc))(callback->context,
+						  plugin_name,
+						  option,
+						  result,
+						  len)
+	      == SASL_OK)
+	    return SASL_OK;
+	}
       }
   }
-
+  
   /* look it up in our configuration file */
   *result = sasl_config_getstring(option, NULL);
   if (*result != NULL) {
@@ -886,7 +1001,7 @@ static int
 _sasl_getsimple(void *context,
 		int id,
 		const char ** result,
-		unsigned * len)
+		size_t *len)
 {
   const char *userid;
   sasl_conn_t *conn;
@@ -931,20 +1046,6 @@ _sasl_getsimple(void *context,
   }
 }
 
-static int
-_sasl_getpath(void *context __attribute__((unused)),
-	      const char **path)
-{
-  if (! path)
-    return SASL_BADPARAM;
-
-  *path = getenv(SASL_PATH_ENV_VAR);
-  if (! *path)
-    *path = PLUGINDIR;
-
-  return SASL_OK;
-}
-
 static int
 _sasl_verifyfile(void *context __attribute__((unused)),
 		 char *file  __attribute__((unused)),
@@ -1091,15 +1192,16 @@ _sasl_log (sasl_conn_t *conn,
 	   ...)
 {
   char *out=(char *) sasl_ALLOC(250);
-  int alloclen=100; /* current allocated length */
-  int outlen=0; /* current length of output buffer */
-  int pos=0; /* current position in format string */
-  int formatlen;
+  size_t alloclen=100; /* current allocated length */
+  size_t outlen=0; /* current length of output buffer */
+  size_t formatlen;
+  size_t pos=0; /* current position in format string */
   int result;
   sasl_log_t *log_cb;
   void *log_ctx;
   
   int ival;
+  unsigned int uval;
   char *cval;
   va_list ap; /* varargs thing */
 
@@ -1204,8 +1306,25 @@ _sasl_log (sasl_conn_t *conn,
 		goto done;
 
 	    done=1;
-
 	    break;
+
+	  case 'o':
+	  case 'u':
+	  case 'x':
+	  case 'X':
+	    frmt[frmtpos++]=fmt[pos];
+	    frmt[frmtpos]=0;
+	    uval = va_arg(ap, unsigned int); /* get the next arg */
+
+	    snprintf(tempbuf,20,frmt,uval); /* have snprintf do the work */
+	    /* now add the string */
+	    result = _sasl_add_string(&out, &alloclen, &outlen, tempbuf);
+	    if (result != SASL_OK)
+		goto done;
+
+	    done=1;
+	    break;
+
 	  default: 
 	    frmt[frmtpos++]=fmt[pos]; /* add to the formating */
 	    frmt[frmtpos]=0;	    
@@ -1220,7 +1339,10 @@ _sasl_log (sasl_conn_t *conn,
     }
   }
 
-  out[outlen]=0; /* put 0 at end */
+  /* put 0 at end */
+  result = _buf_alloc(&out, &alloclen, outlen+1);
+  if (result != SASL_OK) goto done;
+  out[outlen]=0;
 
   va_end(ap);    
 
@@ -1307,12 +1429,12 @@ _sasl_alloc_utils(sasl_conn_t *conn,
   utils->prop_set=&prop_set;
   utils->prop_setvals=&prop_setvals;
   utils->prop_erase=&prop_erase;
+  utils->auxprop_store=&sasl_auxprop_store;
 #endif
 
   /* Spares */
   utils->spare_fptr = NULL;
-  utils->spare_fptr1 = utils->spare_fptr2 = 
-      utils->spare_fptr3 = NULL;
+  utils->spare_fptr1 = utils->spare_fptr2 = NULL;
   
   return utils;
 }
@@ -1401,7 +1523,7 @@ _sasl_find_verifyfile_callback(const sasl_callback_t *callbacks)
 }
 
 /* Basically a conditional call to realloc(), if we need more */
-int _buf_alloc(char **rwbuf, unsigned *curlen, unsigned newlen) 
+int _buf_alloc(char **rwbuf, size_t *curlen, size_t newlen) 
 {
     if(!(*rwbuf)) {
 	*rwbuf = sasl_ALLOC(newlen);
@@ -1430,7 +1552,7 @@ int _buf_alloc(char **rwbuf, unsigned *curlen, unsigned newlen)
 
 /* for the mac os x cfm glue: this lets the calling function
    get pointers to the error buffer without having to touch the sasl_conn_t struct */
-void _sasl_get_errorbuf(sasl_conn_t *conn, char ***bufhdl, unsigned **lenhdl)
+void _sasl_get_errorbuf(sasl_conn_t *conn, char ***bufhdl, size_t **lenhdl)
 {
 	*bufhdl = &conn->error_buf;
 	*lenhdl = &conn->error_buf_len;
@@ -1528,7 +1650,7 @@ int _sasl_ipfromstring(const char *addr,
 	return SASL_BADPARAM;
 
     if (out) {
-	if (outlen < (unsigned)ai->ai_addrlen) {
+	if (outlen < (socklen_t)ai->ai_addrlen) {
 	    freeaddrinfo(ai);
 	    return SASL_BUFOVER;
 	}
@@ -1539,3 +1661,286 @@ int _sasl_ipfromstring(const char *addr,
 
     return SASL_OK;
 }
+
+int _sasl_build_mechlist(void) 
+{
+    int count = 0;
+    sasl_string_list_t *clist = NULL, *slist = NULL, *olist = NULL;
+    sasl_string_list_t *p, *q, **last, *p_next;
+
+    clist = _sasl_client_mechs();
+    slist = _sasl_server_mechs();
+
+    if(!clist) {
+	olist = slist;
+    } else {
+	int flag;
+	
+	/* append slist to clist, and set olist to clist */
+	for(p = slist; p; p = p_next) {
+	    flag = 0;
+	    p_next = p->next;
+
+	    last = &clist;
+	    for(q = clist; q; q = q->next) {
+		if(!strcmp(q->d, p->d)) {
+		    /* They match, set the flag */
+		    flag = 1;
+		    break;
+		}
+		last = &(q->next);
+	    }
+
+	    if(!flag) {
+		*last = p;
+		p->next = NULL;
+	    } else {
+		sasl_FREE(p);
+	    }
+	}
+
+	olist = clist;
+    }
+
+    if(!olist) {
+	printf ("no olist");
+	return SASL_FAIL;
+    }
+
+    for (p = olist; p; p = p->next) count++;
+    
+    if(global_mech_list) {
+	sasl_FREE(global_mech_list);
+	global_mech_list = NULL;
+    }
+    
+    global_mech_list = sasl_ALLOC((count + 1) * sizeof(char *));
+    if(!global_mech_list) return SASL_NOMEM;
+    
+    memset(global_mech_list, 0, (count + 1) * sizeof(char *));
+    
+    count = 0;
+    for (p = olist; p; p = p_next) {
+	p_next = p->next;
+
+	global_mech_list[count++] = (char *) p->d;
+
+    	sasl_FREE(p);
+    }
+
+    return SASL_OK;
+}
+
+const char ** sasl_global_listmech(void) 
+{
+    return (const char **)global_mech_list;
+}
+
+int sasl_listmech(sasl_conn_t *conn,
+		  const char *user,
+		  const char *prefix,
+		  const char *sep,
+		  const char *suffix,
+		  const char **result,
+		  unsigned *plen,
+		  int *pcount)
+{
+    if(!conn) {
+	return SASL_BADPARAM;
+    } else if(conn->type == SASL_CONN_SERVER) {
+	RETURN(conn, _sasl_server_listmech(conn, user, prefix, sep, suffix,
+					   result, plen, pcount));
+    } else if (conn->type == SASL_CONN_CLIENT) {
+	RETURN(conn, _sasl_client_listmech(conn, prefix, sep, suffix,
+					   result, plen, pcount));
+    }
+    
+    PARAMERROR(conn);
+}
+
+
+#ifndef WIN32
+static int
+_sasl_getpath(void *context __attribute__((unused)),
+	      const char **path)
+{
+  if (! path)
+    return SASL_BADPARAM;
+
+  *path = getenv(SASL_PATH_ENV_VAR);
+  if (! *path)
+    *path = PLUGINDIR;
+
+  return SASL_OK;
+}
+
+#else
+/* Return NULL on failure */
+static int
+_sasl_getpath(void *context __attribute__((unused)), const char **path)
+{
+    /* Open registry entry, and find all registered SASL libraries.
+     *
+     * Registry location:
+     *
+     *     SOFTWARE\\Carnegie Mellon\\Project Cyrus\\SASL Library
+     *
+     * Key - value:
+     *
+     *     "SearchPath" - value: PATH like (';' delimited) list
+     *                    of directories where to search for plugins
+     *                    The list may contain references to environment
+     *                    variables (e.g. %PATH%).
+     *
+     */
+    HKEY  hKey;
+    DWORD ret;
+    DWORD ValueType;		    /* value type */
+    DWORD cbData;		    /* value size */
+    BYTE * ValueData;		    /* value */
+    DWORD cbExpandedData;	    /* "expanded" value size */
+    BYTE * ExpandedValueData;	    /* "expanded" value */
+    char * return_value;	    /* function return value */
+    char * tmp;
+
+    /* Initialization */
+    ExpandedValueData = NULL;
+    ValueData = NULL;
+    return_value = NULL;
+
+    /* Open the registry */
+    ret = RegOpenKeyEx(HKEY_LOCAL_MACHINE,
+		       SASL_ROOT_KEY,
+		       0,
+		       KEY_READ,
+		       &hKey);
+
+    if (ret != ERROR_SUCCESS) { 
+		/* no registry entry */
+		*path = PLUGINDIR;
+		return SASL_OK; 
+	}
+
+    /* figure out value type and required buffer size */
+    /* the size will include space for terminating NUL if required */
+    RegQueryValueEx (hKey,
+		     SASL_PATH_SUBKEY,
+		     NULL,	    /* reserved */
+		     &ValueType,
+		     NULL,
+		     &cbData);
+ 
+    /* Only accept string related types */
+    if (ValueType != REG_EXPAND_SZ &&
+	ValueType != REG_MULTI_SZ &&
+	ValueType != REG_SZ) {
+	return_value = NULL;
+	goto CLEANUP;
+    }
+
+    /* Any high water mark? */
+    ValueData = sasl_ALLOC(cbData);
+    if (ValueData == NULL) {
+	return_value = NULL;
+	goto CLEANUP;
+    };
+
+    RegQueryValueEx (hKey,
+		     SASL_PATH_SUBKEY,
+		     NULL,	    /* reserved */
+		     &ValueType,
+		     ValueData,
+		     &cbData);
+
+    switch (ValueType) {
+    case REG_EXPAND_SZ:
+        /* : A random starting guess */
+        cbExpandedData = cbData + 1024;
+        ExpandedValueData = sasl_ALLOC(cbExpandedData);
+        if (ExpandedValueData == NULL) {
+            return_value = NULL;
+            goto CLEANUP;
+        };
+
+        cbExpandedData = ExpandEnvironmentStrings(
+                                                  ValueData,
+                                                  ExpandedValueData,
+                                                  cbExpandedData);
+
+        if (cbExpandedData == 0) {
+            /* : GetLastError() contains the reason for failure */
+            return_value = NULL;
+            goto CLEANUP;
+        }
+
+        /* : Must retry expansion with the bigger buffer */
+        if (cbExpandedData > cbData + 1024) {
+            /* : Memory leak here if can't realloc */
+            ExpandedValueData = sasl_REALLOC(ExpandedValueData, cbExpandedData);
+            if (ExpandedValueData == NULL) {
+                return_value = NULL;
+                goto CLEANUP;
+            };
+
+            cbExpandedData = ExpandEnvironmentStrings(
+                                                      ValueData,
+                                                      ExpandedValueData,
+                                                      cbExpandedData);
+
+            /* : This should not happen */
+            if (cbExpandedData == 0) {
+                /* : GetLastError() contains the reason for failure */
+                return_value = NULL;
+                goto CLEANUP;
+            }
+        }
+
+        sasl_FREE(ValueData);
+        ValueData = ExpandedValueData;
+        /* : This is to prevent automatical freeing of this block on cleanup */
+        ExpandedValueData = NULL;
+
+        break;
+
+    case REG_MULTI_SZ:
+        tmp = ValueData;
+
+        /* : We shouldn't overflow here, as the buffer is guarantied
+           : to contain at least two consequent NULs */
+        while (1) {
+            if (tmp[0] == '\0') {
+                /* : Stop the process if we found the end of the string (two consequent NULs) */
+                if (tmp[1] == '\0') {
+                    break;
+                }
+
+                /* : Replace delimiting NUL with our delimiter characted */
+                tmp[0] = PATHS_DELIMITER;
+            }
+            tmp += strlen(tmp);
+        }
+        break;
+
+    case REG_SZ:
+        /* Do nothing, it is good as is */
+        break;
+
+    default:
+        return_value = NULL;
+        goto CLEANUP;
+    }
+
+    return_value = ValueData;
+
+    CLEANUP:
+    RegCloseKey(hKey);
+    if (ExpandedValueData != NULL) sasl_FREE(ExpandedValueData);
+    if (return_value == NULL) {
+	if (ValueData != NULL) sasl_FREE(ValueData);
+    }
+    *path = return_value;
+
+	return SASL_OK;
+}
+
+#endif
